/***************************************************************************
                          matrix.cpp  -  description
                             -------------------
    begin                : Thu Aug 8 2002
    copyright            : (C) 2002 by Nobuhiro Kuroiwa
    email                : nobu@iastate.edu
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "matrix.h"

Error::Error( string newmsg)
{
	error_message = newmsg;
}
Error::Error( char* newmsg)
{
	error_message.assign( newmsg);
}

//---------------------------------------------------------------------------
//	Matrix::Matrix()
//---------------------------------------------------------------------------
Matrix::Matrix() :
raw_size( 3), col_size( 1)
{
	init();
}

//---------------------------------------------------------------------------
//	Matrix::Matrix()
//---------------------------------------------------------------------------
Matrix::Matrix( unsigned int raw_in, unsigned int col_in)
:
raw_size( 3), col_size( 1)
{
	if( raw_in != 0)	raw_size = raw_in;
	if( col_in != 0)	col_size = col_in;
	init();
}

//---------------------------------------------------------------------------
//	Matrix::getVal()
//---------------------------------------------------------------------------
NUM_TYPE Matrix::getVal( unsigned int raw_in, unsigned int col_in)
{
	if( !checkBound( raw_in, col_in))
		throw Error( "Out of Boundary");
	return vec[ raw_in][ col_in];
}
//---------------------------------------------------------------------------
//	Matrix::init()
//---------------------------------------------------------------------------
void Matrix::init()
{
	vector< NUM_TYPE> vec_col( col_size, 0.0);
	unsigned int counter;
	for( counter = 0; counter < col_size; counter++)
		vec_col.push_back( 0.0);

	for( counter = 0; counter < raw_size; counter++)
		vec.push_back( vec_col);
}

//---------------------------------------------------------------------------
//	Matrix::checkBound()
//---------------------------------------------------------------------------
bool Matrix::checkBound( unsigned int raw_in, unsigned int col_in)
{
	if( raw_in < 0 || raw_in >= raw_size)
		return false;
	if( col_in < 0 || col_in >= col_size)
		return false;
	return true;	
}

//---------------------------------------------------------------------------
//	Matrix::assign()
//---------------------------------------------------------------------------
void Matrix::assign(
	 unsigned int raw_in, unsigned int col_in, NUM_TYPE new_val
)
{
	if( !checkBound( raw_in, col_in))
	{
		throw Error( "Out of Boundary");
	}
	vec[ raw_in][ col_in] = new_val;
}

//---------------------------------------------------------------------------
//	Matrix::multiple()
//---------------------------------------------------------------------------
Matrix Matrix::multiple( Matrix &mat_in)
{
	unsigned int l, m, n;
	unsigned int counter_i, counter_j, counter_k;

	l = raw_size;
	m = col_size;
	n = mat_in.colSize();

	Matrix mat_ret( l, n);

	for( counter_i = 0; counter_i < l; counter_i++)
		for( counter_k = 0; counter_k < n; counter_k++)
			for( counter_j = 0; counter_j < m; counter_j++)
				mat_ret.assign(
								counter_i,
								counter_k,
								mat_ret.getVal( counter_i, counter_k)
								+ vec[ counter_i][ counter_j]
								* mat_in.getVal( counter_j, counter_k)
							);
	return mat_ret;
}

//---------------------------------------------------------------------------
//	Matrix::multiple()
//---------------------------------------------------------------------------
Matrix Matrix::multiple( Matrix *pmat_in)
{
	Matrix mat_temp = *pmat_in;
	return multiple( mat_temp);
}

//---------------------------------------------------------------------------
//	Matrix::add()
//---------------------------------------------------------------------------
Matrix Matrix::add( Matrix &mat_in)
{
	unsigned int counter_in, counter_out;
	Matrix ret_mat( raw_size, col_size);
	for( counter_out = 0; counter_out < raw_size; counter_out++)
	{
		for( counter_in = 0; counter_in < col_size; counter_in++)
		{
			ret_mat.assign(
				counter_out,
				counter_in,
				vec[ counter_out][ counter_in]
			 	+ mat_in.getVal( counter_out, counter_in)
			);
		}
	}
	return ret_mat;
}

//---------------------------------------------------------------------------
//	Matrix::rawSize()
//---------------------------------------------------------------------------
unsigned int Matrix::rawSize() const
{
	return raw_size;
}

//---------------------------------------------------------------------------
//	Matrix::colSize()
//---------------------------------------------------------------------------
unsigned int Matrix::colSize() const
{
	return col_size;
}

//---------------------------------------------------------------------------
//	Matrix::printMatrix()
//---------------------------------------------------------------------------
void Matrix::printMatrix()
{
     unsigned int counter_in, counter_out;
	for( counter_out = 0; counter_out < raw_size; counter_out++)
	{
		for( counter_in = 0; counter_in < col_size; counter_in++)
		{
			cout << "( " << counter_out << ", " << counter_in << ") = ";
			cout << vec[ counter_out][counter_in] << "\t";
		}
		cout << "\n";
	}
	
}

//---------------------------------------------------------------------------
//	Vec::Vec()
//---------------------------------------------------------------------------
Vec::Vec()
{
}

//---------------------------------------------------------------------------
//	Vec::Vec()
//---------------------------------------------------------------------------
Vec::Vec( NUM_TYPE new_x, NUM_TYPE new_y, NUM_TYPE new_z)
{
	Vec::assign( new_x, new_y, new_z);
}

//---------------------------------------------------------------------------
//	Vec::x()
//---------------------------------------------------------------------------
NUM_TYPE Vec::x()
{
	return vec[0][0];
}

//---------------------------------------------------------------------------
//	Vec::y()
//---------------------------------------------------------------------------
NUM_TYPE Vec::y()
{
	return vec[1][0];
}

//---------------------------------------------------------------------------
//	Vec::z()
//---------------------------------------------------------------------------
NUM_TYPE Vec::z()
{
	return vec[2][0];
}

//---------------------------------------------------------------------------
//	Vec::x()
//---------------------------------------------------------------------------
void Vec::x( NUM_TYPE new_val)
{
	vec[0][0] = new_val;
}

//---------------------------------------------------------------------------
//	Vec::y()
//---------------------------------------------------------------------------
void Vec::y( NUM_TYPE new_val)
{
	vec[1][0] = new_val;
}

//---------------------------------------------------------------------------
//	Vec::z()
//---------------------------------------------------------------------------
void Vec::z( NUM_TYPE new_val)
{
	vec[2][0] = new_val;
}

//---------------------------------------------------------------------------
//	Vec::multiple()
//---------------------------------------------------------------------------
Matrix	Vec::multiple( Vec &vec_in)
{
	Matrix mat = vec_in;
	return Matrix::multiple( mat);
}

//---------------------------------------------------------------------------
//	Vec::add()
//---------------------------------------------------------------------------
Vec Vec::add( Vec &vec_in)
{
	Matrix mat_temp	= vec_in;
	Matrix mat_ans	= Matrix::add( mat_temp);
	Vec vec_ret;
	vec_ret.assign( mat_ans);
	return vec_ret;
}

//---------------------------------------------------------------------------
//	Vec::assign()
//---------------------------------------------------------------------------
void Vec::assign( Matrix &mat_in)
{
	vec[0][0] = mat_in.getVal( 0, 0);
	vec[1][0] = mat_in.getVal( 1, 0);
	vec[2][0] = mat_in.getVal( 2, 0);
}

//---------------------------------------------------------------------------
//	Vec::assign()
//---------------------------------------------------------------------------
void Vec::assign( NUM_TYPE new_x, NUM_TYPE new_y, NUM_TYPE new_z)
{
	vec[0][0] = new_x;
	vec[1][0] = new_y;
	vec[2][0] = new_z;
}

//---------------------------------------------------------------------------
//	rotateMatrix_Z()
//---------------------------------------------------------------------------
Matrix rotateMatrix_Z( double angle)
{
	Matrix mat_ret( 3, 3);
	double	cos_temp = cos( angle);
	double	sin_temp = sin( angle);

	mat_ret.assign( 0, 0, cos_temp);
	mat_ret.assign( 0, 1, sin_temp);
	mat_ret.assign( 0, 2, 0);

	mat_ret.assign( 1, 0, -sin_temp);	
	mat_ret.assign( 1, 1, cos_temp);
	mat_ret.assign( 1, 2, 0);

	mat_ret.assign( 2, 0, 0);	
	mat_ret.assign( 2, 1, 0);
	mat_ret.assign( 2, 2, 1);

	return mat_ret;
}

//---------------------------------------------------------------------------
//	rotateMatrix_Y()
//---------------------------------------------------------------------------
Matrix rotateMatrix_Y( double angle)
{
	Matrix mat_ret( 3, 3);
	double	cos_temp = cos( angle);
	double	sin_temp = sin( angle);

	mat_ret.assign( 0, 0, cos_temp);
	mat_ret.assign( 0, 1, 0);
	mat_ret.assign( 0, 2, -sin_temp);

	mat_ret.assign( 1, 0, 0);	
	mat_ret.assign( 1, 1, 1);
	mat_ret.assign( 1, 2, 0);

	mat_ret.assign( 2, 0, sin_temp);	
	mat_ret.assign( 2, 1, 0);
	mat_ret.assign( 2, 2, cos_temp);

	return mat_ret;
}

//---------------------------------------------------------------------------
//	rotateMatrix_X()
//---------------------------------------------------------------------------
Matrix rotateMatrix_X( double angle)
{
	Matrix mat_ret( 3, 3);
	double	cos_temp = cos( angle);
	double	sin_temp = sin( angle);

	mat_ret.assign( 0, 0, 1);
	mat_ret.assign( 0, 1, 0);
	mat_ret.assign( 0, 2, 0);

	mat_ret.assign( 1, 0, 0);	
	mat_ret.assign( 1, 1, cos_temp);
	mat_ret.assign( 1, 2, sin_temp);

	mat_ret.assign( 2, 0, 0);	
	mat_ret.assign( 2, 1, -sin_temp);
	mat_ret.assign( 2, 2, cos_temp);

	return mat_ret;
}

//---------------------------------------------------------------------------
//	round()
//  Round number
//  Looks like math.h has same function. So tehmporarily disactivated
//	for KDE.
//---------------------------------------------------------------------------
#ifndef MSVC
int round( NUM_TYPE num_in)
{
	int factor = 1;
	if( num_in < 0.0)	factor = -1;
	NUM_TYPE num_ret = fabs( num_in);
	return int( NUM_TYPE( factor) * ( num_ret + 0.5));
}
#endif

//---------------------------------------------------------------------------
//	roundUp()
//---------------------------------------------------------------------------
int roundUp( NUM_TYPE num_in)
{
	int factor = 1;
	if( num_in < 0.0)	factor = -1;
	if( int( fabs( num_in)) < fabs( double( num_in)))
		return	int( num_in + factor * 1);
	return int( num_in);  
}

//---------------------------------------------------------------------------
//	vecLength()
//---------------------------------------------------------------------------
NUM_TYPE vecLength( Matrix &mat_in)
{
	return sqrt(
					pow( mat_in.getVal( 0, 0), 2)
					+ pow( mat_in.getVal( 1, 0), 2)
					+ pow( mat_in.getVal( 2, 0), 2)
				);
}

//---------------------------------------------------------------------------
//	dotProduct()
//---------------------------------------------------------------------------
NUM_TYPE dotProduct( Matrix &mat_a, Matrix &mat_b)
{
	return
			mat_a.getVal( 0, 0) * mat_b.getVal( 0, 0)
			+ mat_a.getVal( 1, 0) * mat_b.getVal( 1, 0)
			+ mat_a.getVal( 2, 0) * mat_b.getVal( 2, 0);
}

