/***************************************************************************
                          matrix.h  -  description
                             -------------------
    begin                : Thu Aug 8 2002
    copyright            : (C) 2002 by Nobuhiro Kuroiwa
    email                : nobu@iastate.edu
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <iostream>
#include <vector>
#include <string>
#include <cmath>
#ifndef MATRIX_H
#define	MATRIX_H

#define		NUM_TYPE	double
#ifndef PI
	#define PI 3.1414926535
#endif

using namespace std;

typedef std::vector <NUM_TYPE> VecW;

class Error
{
public:
	Error( string newmsg);
	Error( char* newmsg);
	string error_message;
};

class Matrix
{
public:
	Matrix();
	Matrix( unsigned int raw_in, unsigned int col_in);
	NUM_TYPE	getVal( unsigned int raw_in, unsigned int col_in);
	void		assign(
						unsigned int raw_in, 
						unsigned int col_in, 
						NUM_TYPE new_val
				);
	bool		checkBound( unsigned int raw_in, unsigned int col_in);
	Matrix		multiple( Matrix &mat_in);
	Matrix		multiple( Matrix *pmat_in);
	Matrix		add( Matrix &mat_in);
	unsigned int	colSize() const;
	unsigned int	rawSize() const;
	void		printMatrix();
protected:
	vector < VecW> vec;
	unsigned int raw_size;
	unsigned int col_size;
	void init();
};

class Vec : public Matrix
{
public:
	Vec();
	Vec( NUM_TYPE new_x, NUM_TYPE new_y, NUM_TYPE new_z);
	NUM_TYPE	x();
	NUM_TYPE	y();
	NUM_TYPE	z();
	void		x( NUM_TYPE new_val);
	void		y( NUM_TYPE new_val);
	void		z( NUM_TYPE new_val);
	Matrix	multiple( Vec &vec_in);
	Vec			add( Vec &vec_in);
	void		assign( Matrix &mat_in);
	void		assign( NUM_TYPE new_x, NUM_TYPE new_y, NUM_TYPE new_z);
};

Matrix rotateMatrix_Z( double angle);
Matrix rotateMatrix_Y( double angle);
Matrix rotateMatrix_X( double angle);
#ifndef MSVC
	int round(    NUM_TYPE num_in);
#endif
int roundUp(  NUM_TYPE num_in);
NUM_TYPE vecLength(   Matrix &vec_in);
NUM_TYPE dotProduct(  Matrix &vec_a, Matrix &vec_b);

#endif

