/***************************************************************************
                          main.cpp  -  description
                             -------------------
    begin                : Wed Aug  7 17:00:58 CDT 2002
    copyright            : (C) 2002 by Nobuhiro Kuroiwa
    email                : nobu@iastate.edu
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "projtool.h"

unsigned int	HEIGHT_TRIANGLE;	//=	100.0 * sqrt( 3) / 2.0;
double			  RESOLUTION;			//=	1.0 / 100.0;

//---------------------------------------------------------------------------
//	initPlanet()
//---------------------------------------------------------------------------
void initPlanet(
					unsigned	int	width_triangle,
					unsigned	int	&height_triangle,
					double		resolution)
{
	RESOLUTION		  =	resolution;
	HEIGHT_TRIANGLE	=	round( double( width_triangle) * sqrt( 3.0) / 2.0);
	height_triangle	=	HEIGHT_TRIANGLE;
}

//---------------------------------------------------------------------------
//	Square::Square()
//---------------------------------------------------------------------------
Square::Square( Vec vec1, Vec vec2, Vec vec3, Vec vec4)
{
	Square::assign( vec1, vec2, vec3, vec4);
}

//---------------------------------------------------------------------------
//	Square::Square()
//---------------------------------------------------------------------------
Square::Square(){}

//---------------------------------------------------------------------------
//	Square::Square()
//---------------------------------------------------------------------------
void Square::assign( Vec vec1, Vec vec2, Vec vec3, Vec vec4)
{
	vec[0]	= vec1;
	vec[1]	= vec2;
	vec[2]	= vec3;
	vec[3]	= vec4;
}

//---------------------------------------------------------------------------
//	Square::Square()
//---------------------------------------------------------------------------
bool Square::iassign(
						unsigned int	index,
						NUM_TYPE 		newx,
						NUM_TYPE		newy,
						NUM_TYPE		newz
)
{
	if( index < 0 || index >= 4)
		return false;

	vec[ index].x( newx);
	vec[ index].y( newy);
	vec[ index].z( newz);

	return true;
}

//---------------------------------------------------------------------------
//	Square::Square()
//---------------------------------------------------------------------------
Vec & Square::getVec( unsigned int index)
{
	if( index < 0 || index >= 4)
		throw Error( "Out of Boundary");
	return vec[ index];
}

//---------------------------------------------------------------------------
//	Square::Square()
//---------------------------------------------------------------------------
Square	Square::add( NUM_TYPE in)
{
	return this->add( in, in, in);
}

//---------------------------------------------------------------------------
//	Square::Square()
//---------------------------------------------------------------------------
bool Square::assign( unsigned int index, Vec &vec_in)
{
	if( index < 0 || index >= 4)
		return false;
	vec[ index] = vec_in;
	return true;
}

//---------------------------------------------------------------------------
//	Square::Square()
//---------------------------------------------------------------------------
Square	Square::add( NUM_TYPE in_x, NUM_TYPE in_y, NUM_TYPE in_z)
{
	Square sq_ret;
	unsigned int counter;
	for( counter = 0; counter < 4; counter++)
		sq_ret.iassign(
						counter,
						vec[counter].x() + in_x,
						vec[counter].y() + in_y,
						vec[counter].z() + in_z
					);
	return sq_ret;
}

//---------------------------------------------------------------------------
//	Square::multiple()
//---------------------------------------------------------------------------
Square	Square::multiple( Matrix &mat_in)
{
	Square			sq_ret;
	unsigned int	counter;
	Matrix			mat_ret;
	Vec				vec_ret;

	for( counter = 0; counter < 4; counter++)
	{
		mat_ret = mat_in.multiple( vec[ counter]);
		vec_ret.assign( mat_ret);
		sq_ret.assign( counter, vec_ret);
	}
	return sq_ret;
} 

//---------------------------------------------------------------------------
//						getSquareOnTriangle()
//
//	Usage of This Function
//	for( y < sqrt(3)/2, y+=RESO)
//		for( x < 1-sqrt(3)y/3, x += RESO)
//			get square x0, y0, width = height = RESO
//---------------------------------------------------------------------------
bool getSquareOnTriangle( NUM_TYPE x0, NUM_TYPE y0, Square &sq)
{
	if(
		(	x0 < 0.0						)
		||
		(   x0 > 1.0 - y0 * sqrt( 3.0) / 3.0	)
	)
		return false;
	if(
		( y0 < 0.0							)
		||                   
		( y0 > sqrt( 3.0) / 2.0				)
		)
		return false;

	sq.iassign( 3, x0,				y0,					0.0);	
	sq.iassign( 0, x0 + RESOLUTION,	y0,					0.0);
	sq.iassign( 1, x0 + RESOLUTION,	y0 + RESOLUTION,	0.0);
	sq.iassign( 2, x0,				y0 + RESOLUTION,	0.0);
	return true;
}

//---------------------------------------------------------------------------
//	sqcnv2p1()
//---------------------------------------------------------------------------
Square sqcnv2p1( Square &sq_in)
{
	Vec vec_ret[4];
	Vec vec_temp;
	Square sq_ret;
	for( unsigned int counter = 0; counter < 4; counter++)
	{
		vec_temp = sq_in.getVec( counter);
		vec_ret[ counter] = cnv2p1( vec_temp);
	}
	sq_ret.assign( vec_ret[0], vec_ret[1], vec_ret[2], vec_ret[3]);
	return sq_ret;
}

//---------------------------------------------------------------------------
//	sqcnv2p2fromp1()
//---------------------------------------------------------------------------
Square sqcnv2p2fromp1( Square &sq_in)
{
	Vec vec_a, vec_b, vec_c, vec_d;
	Square sq_ret;

	vec_a = cnv2p2fromp1( sq_in.getVec( 0));
	vec_b = cnv2p2fromp1( sq_in.getVec( 1));
	vec_c = cnv2p2fromp1( sq_in.getVec( 2));
	vec_d = cnv2p2fromp1( sq_in.getVec( 3));

	sq_ret.assign( vec_a, vec_b, vec_c, vec_d);

	return sq_ret;		
}

//---------------------------------------------------------------------------
//	sqcnv2p3fromp1()
//---------------------------------------------------------------------------
Square sqcnv2p3fromp1( Square &sq_in)
{
	Vec vec_a, vec_b, vec_c, vec_d;
	Square sq_ret;

	vec_a = cnv2p3fromp1( sq_in.getVec( 0));
	vec_b = cnv2p3fromp1( sq_in.getVec( 1));
	vec_c = cnv2p3fromp1( sq_in.getVec( 2));
	vec_d = cnv2p3fromp1( sq_in.getVec( 3));

	sq_ret.assign( vec_a, vec_b, vec_c, vec_d);

	return sq_ret;
}

//---------------------------------------------------------------------------
//	sqcnv2p4()
//---------------------------------------------------------------------------
Square sqcnv2p4( Square &sq_in)
{
	Vec vec_a, vec_b, vec_c, vec_d;
	Square sq_ret;

	vec_a = cnv2p4( sq_in.getVec( 0));
	vec_b = cnv2p4( sq_in.getVec( 1));
	vec_c = cnv2p4( sq_in.getVec( 2));
	vec_d = cnv2p4( sq_in.getVec( 3));

	sq_ret.assign( vec_a, vec_b, vec_c, vec_d);

	return sq_ret;
}

//---------------------------------------------------------------------------
//	cnv2p1()
//---------------------------------------------------------------------------
Vec cnv2p1( Vec& vec_in)
{
	Vec vec_ret;
	Vec vec_add;
	Matrix mat_temp;
	static double	const	PIov6 = PI/6.0;
	static double	const	cosPIov6 = cos( PIov6);
	static double	const	factor_l	= 1.0 / ( 2.0 * cosPIov6);
	static double	const	negFactor_lxSinPIov6 = -factor_l * sin( PIov6);
	//static double negFactor_lxSinPIov6 = -factor_l * sin( PI/6.0);
	static double	const	factor_lxCosPIov6 = factor_l * cosPIov6;
	static double	const	negSqrt6ov3ov3 = -sqrt( 6.0) / 3.0 / 3.0;
	static Matrix	const	mat_X = rotateMatrix_X( PI + acos( 1.0/3.0));
	//-----------------------------------------------------------------------
	// Rotate PI+arccos(1/3) around X axis
	//-----------------------------------------------------------------------
	Matrix mat_rotation = mat_X;
	mat_temp = mat_rotation.multiple( vec_in);
	//-----------------------------------------------------------------------
	// Rotate PI/6 around Z axis
	//-----------------------------------------------------------------------
    mat_rotation	= rotateMatrix_Z( PIov6);
    mat_temp		= mat_rotation.multiple( mat_temp);
    vec_ret.assign( mat_temp);
	//-----------------------------------------------------------------------
	// Move
	//factor_l	:= 1 / ( 2 * cos( PI/6));
	//mat_addup	:= dommat([
	//				[-factor_l * sin( PI/6)],
	//				[ factor_l * cos( PI/6)],
	//				[ 0				]
	//				]);
	//-----------------------------------------------------------------------
	vec_add.x( negFactor_lxSinPIov6		); 
	vec_add.y( factor_lxCosPIov6	 	);
	vec_add.z( 0						);
	vec_ret = vec_ret.add( vec_add);

	// Move
	//.mat_addup	:= dommat([
	//				[ 0],
	//				[ 0],
	//				[ -sqrt(6)/3/3]
	//				]);
	vec_add.x( 0						);
	vec_add.y( 0						);
	vec_add.z( negSqrt6ov3ov3			);
	vec_ret	= vec_ret.add( vec_add);

	return vec_ret;
}

//---------------------------------------------------------------------------
//	cnv2p2fromp1()
//---------------------------------------------------------------------------
Vec cnv2p2fromp1( Vec &vec_p1)
{
	Vec				vec_ret;
	Matrix			mat_temp;
	static	Matrix	mat_rotation = rotateMatrix_Z( 2.0 * PI / 3.0);
	mat_temp = mat_rotation.multiple( vec_p1);
	vec_ret.assign( mat_temp);

	return vec_ret;
}

//---------------------------------------------------------------------------
//	cnv2p3fromp1()
//---------------------------------------------------------------------------
Vec cnv2p3fromp1( Vec &vec_p1)
{
	Vec				vec_ret;
	Matrix			mat_temp; 
	static	Matrix	mat_rotation = rotateMatrix_Z( - 2.0 * PI / 3.0);
	mat_temp = mat_rotation.multiple( vec_p1);
	vec_ret.assign( mat_temp);

	return vec_ret;
}

//---------------------------------------------------------------------------
//	cnv2p4()
//---------------------------------------------------------------------------
Vec cnv2p4( Vec &vec_in)
{


	Vec		vec_ret;
	Matrix	mat_rotation( 3, 3);
	Vec		vec_add;
	static const		Matrix	mat_Z = rotateMatrix_Z( -PI/6.0);
	static double const	negSqrt3ov6 = -sqrt( 3.0)/6.0;
	static double const	neg1ov2 = -1.0/2.0;
	static double const	negSqrt6ov3ov3 = -sqrt( 6.0) / 3.0 / 3.0;
	//-----------------------------------------------------------------------
	// Rotate -PI/6 around Z axis
	//-----------------------------------------------------------------------
	mat_rotation = mat_Z;
	Matrix mat_temp = mat_rotation.multiple( vec_in);
	vec_ret.assign( mat_temp);
	//-----------------------------------------------------------------------
	// move the vector
	//mat_addup	:= dommat( [ -sqrt(3)/6, -1/2, 0]);
	//-----------------------------------------------------------------------
	vec_add.assign( negSqrt3ov6, neg1ov2, 0);
	vec_ret = vec_ret.add( vec_add);	
	//-----------------------------------------------------------------------
	//mat_addup	:= dommat([
	//				[ 0],
	//				[ 0],
	//				[ -sqrt(6)/3/3]
	//				]);
	//-----------------------------------------------------------------------
	vec_add.assign( 0, 0, negSqrt6ov3ov3);
	vec_ret = vec_ret.add( vec_add);
	
	return vec_ret;
}

//---------------------------------------------------------------------------
//	cartesian2Spherical()
//---------------------------------------------------------------------------
Vec cartesian2Spherical( Vec &vec_in)
{
	double f_alpha	= 1.0;

    double	len_projxy, len_unitx, cosalpha, ret_alpha, ret_beta, rad;
	Vec		ret_vec;
	//-----------------------------------------------------------------------
	//  Unit Vector of along X axis
	//	unitx	:=	dommat( [ [1], [0], [0]]);
	//-----------------------------------------------------------------------
	Vec unitx;
	unitx.assign( 1, 0, 0);
	//-----------------------------------------------------------------------
	// projection of mat_in onto x-y plane	
	//projxy	:=	dommat( [ [ in_x], [ in_y], [0]]);
	//-----------------------------------------------------------------------
	Vec projxy;
	projxy = vec_in;
	projxy.z( 0);
	//-----------------------------------------------------------------------
	// length of projxy
	//-----------------------------------------------------------------------
	len_projxy	= vecLength( projxy);
	len_unitx	= 1.0; 
	cosalpha	= dotProduct( unitx, projxy) / ( len_projxy * len_unitx);
	ret_alpha	= acos( cosalpha);
	//-----------------------------------------------------------------------
	// if y is negative
	//-----------------------------------------------------------------------
	if( vec_in.y() > 0.0)
		f_alpha	= -1.0;
	
	rad			= vecLength( vec_in);
	ret_beta	= acos( vec_in.z() / rad );
	//-----------------------------------------------------------------------
	//ret_mat		= dommat(
	//					[
	//						[ ret_alpha	* f_alpha	],
	//						[ ret_beta				],
	//						[ 1						]
	//					]
	//				);
	//-----------------------------------------------------------------------
	ret_vec.x( ret_alpha * f_alpha);
	ret_vec.y( ret_beta);
	ret_vec.z( 1);

	return ret_vec;
}

//---------------------------------------------------------------------------
//	spherical2Mercator()
//---------------------------------------------------------------------------
Vec spherical2Mercator( Vec &vec_in)
{
	Vec ret_vec;
	static const double	PIov2 = PI / 2.0;
	ret_vec.x( vec_in.x());
	ret_vec.y( sin( PIov2 - vec_in.y()));
	ret_vec.z( 0);
	return ret_vec;
}

//---------------------------------------------------------------------------
//	sqcartesian2Spherical()
//---------------------------------------------------------------------------
Square sqcartesian2Spherical( Square &sq_in)
{
	Square sq_ret;
	Vec vec_ret[4];
	Vec vec_temp;

	for( unsigned int counter = 0; counter < 4; counter++)
	{
		vec_temp = sq_in.getVec( counter);
		vec_ret[ counter] = cartesian2Spherical( vec_temp);
	}
	sq_ret.assign( vec_ret[0], vec_ret[1], vec_ret[2], vec_ret[3]);
	return sq_ret;
}

//---------------------------------------------------------------------------
//	sqspherical2Mercator()
//---------------------------------------------------------------------------
Square sqspherical2Mercator( Square &sq_in)
{
	Square sq_ret;
	Vec vec_temp[4];
	for( unsigned int counter = 0; counter < 4; counter++)
	{
		vec_temp[ counter] = spherical2Mercator( sq_in.getVec( counter));
	}
	sq_ret.assign( vec_temp[0], vec_temp[1], vec_temp[2], vec_temp[3]);
	return sq_ret;
}

//---------------------------------------------------------------------------
///	getLarger()
//---------------------------------------------------------------------------
NUM_TYPE getLarger( NUM_TYPE in_a, NUM_TYPE in_b)
{
	if( in_a > in_b)	return in_a;
	return in_b;
}

//---------------------------------------------------------------------------
//	getSmaller
//---------------------------------------------------------------------------
NUM_TYPE getSmaller( NUM_TYPE in_a, NUM_TYPE in_b)
{
	if( in_a < in_b)	return in_a;
	return in_b;
}

//---------------------------------------------------------------------------
//	getOuterSquare()
//---------------------------------------------------------------------------
Square getOuterSquare( Square &sq_in)
{
	int counter;
	Vec leftup, rightdown, vectemp, leftdown, rightup;
	NUM_TYPE xmin, xmax, ymin, ymax;
	Square sq_ret;

	vectemp = sq_in.getVec( 0);
	xmin = vectemp.x();
	xmax = vectemp.x();
	ymin = vectemp.y();
	ymax = vectemp.y();
	
	for( counter = 1; counter < 4; counter++)
	{
		vectemp = sq_in.getVec( counter);
		if( xmin > vectemp.x())		xmin = vectemp.x();
		if( xmax < vectemp.x())		xmax = vectemp.x();
		if( ymin > vectemp.y())		ymin = vectemp.y();
		if( ymax < vectemp.y())		ymax = vectemp.y();
	}

	leftdown.assign(	xmin,	 ymin,		0						);
	leftup.assign(		xmin,	 ymax,		0						);
	rightup.assign(		xmax,	 ymax,		0						);
	rightdown.assign(	xmax,	 ymin,		0						);
	sq_ret.assign(		leftup,	rightup,	rightdown,	leftdown	);

	return sq_ret;	
}

//---------------------------------------------------------------------------
//	fixPos()
//	Fix the input value between 0 and width
//---------------------------------------------------------------------------
unsigned int fixPos( unsigned int i_in, unsigned int width)
{
	unsigned int ret;
	if( i_in < 0)			ret = 0;
	else if( i_in >= width)	ret = width - 1;
	else					ret = i_in;

	return ret;
}



