/***************************************************************************
                          sankaku.cpp  -  description
                             -------------------
    begin                : Thu Nov 21 2002
    copyright            : (C) 2002 by Nobuhiro Kuroiwa
    email                : nobu@iastate.edu
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   any later version.    					                               *
 *                                                                         *
 ***************************************************************************/
#include "sankaku.h"
#include "paragui/pgdraw.h"
#include "stdlib.h"
#include "SDL/SDL_image.h"
//---------------------------------------------------------------------------
//	 SankakuCallBack::SankakuCallBack()
//---------------------------------------------------------------------------
SankakuCallBack::SankakuCallBack(){}

//---------------------------------------------------------------------------
//	 SankakuCallBack::~SankakuCallBack()
//---------------------------------------------------------------------------
SankakuCallBack::~SankakuCallBack(){}

//---------------------------------------------------------------------------
//	Sankaku::Sankaku()
//---------------------------------------------------------------------------
Sankaku::Sankaku( string *pstr_filename2load)
{
	stop			= false;
	ready2play		= false;
	pcallback		= NULL;
	sfc_mercator	= NULL;
	sfc_progress	= NULL;
	sfc_result		= NULL;
	sfc_temp_unfold = NULL;
	for( int i = 0; i < 4; i++)	sfc_trig_array[ i] = NULL;
	nullEmAllButMercator();
	if( !load( pstr_filename2load))
		cerr << "loading file failed" << endl;
}

//---------------------------------------------------------------------------
//	Sankaku::~Sankaku()
//---------------------------------------------------------------------------
Sankaku::~Sankaku()
{
	freeEmAllButMercator();
	freeSurface( sfc_mercator);
}

//---------------------------------------------------------------------------
//	Sankaku::nullEmAllButMercator()
//---------------------------------------------------------------------------
void Sankaku::nullEmAllButMercator()
{
	sfc_progress	= NULL;
	sfc_result		= NULL;
	sfc_temp_unfold = NULL;
	for( unsigned short counter = 0; counter < 4; counter++)
		sfc_trig_array[ counter] = NULL;
}

//---------------------------------------------------------------------------
//	Sankaku::freeEmAllButMercator()
//---------------------------------------------------------------------------
void Sankaku::freeEmAllButMercator()
{
	for( unsigned short counter = 0; counter < 4; counter++)
		freeSurface( sfc_trig_array[ counter]);
    freeSurface( sfc_progress	);
	freeSurface( sfc_result		);
	freeSurface( sfc_temp_unfold);	
}

//---------------------------------------------------------------------------
//	Sankaku::setCallBack()
//	Assign the call back function
//---------------------------------------------------------------------------
void Sankaku::setCallBack( SankakuCallBack *pclient)
{
	pcallback = pclient;
}

//---------------------------------------------------------------------------
//	 Sankaku::isReady()
//---------------------------------------------------------------------------
bool Sankaku::isReady()		const
{
	return ready2play;
}

//---------------------------------------------------------------------------
//	 Sankaku::getTrigIndex()
//---------------------------------------------------------------------------
int Sankaku::getTrigIndex() const
{
	return current_trig;
}




//---------------------------------------------------------------------------
//	 Sankaku::stopIt()
//---------------------------------------------------------------------------
void Sankaku::stopIt()
{
	stop = true;
}

//---------------------------------------------------------------------------
//	createSurface()
//---------------------------------------------------------------------------
bool Sankaku::createSurface(
					SDL_Surface*	&sfc_in,
					unsigned int	width,
					unsigned int	height,
					bool			flg_clear
)
{
	sfc_in = SDL_CreateRGBSurface(
										SDL_SWSURFACE,
										width,
										height,
										32,
										0,0,0,0
										//0, 0, 0, 0	// black
									);
	if( sfc_in == NULL) return false;
	if( flg_clear)      sge_ClearSurface( sfc_in, 
												//255, 255, 255);
												0x00, 0x00, 0x00);

	return true;
}


//---------------------------------------------------------------------------
//	getBoxOnMercator()
//
//	Obtain number of box formed by sq_in and the theoretical coordinate in
//	psq_array_ret[].
//	If sq_in cross edge of 2*2PI plane,
//	it forms two boxes. In this case return 2 and set theoretical coordinates
//	in the psq_array_ret[] for each box.
//	if multiple box obtained return number of the boxes
//	This Function doesn't treat y value. That's must be done someday...
//
//	sq_in			: Coord. of square which may or may not cross the edge
//	psq_array_ret	: Coordinates of squares guaranteed not to cross the edge

//---------------------------------------------------------------------------
unsigned int Sankaku::getBoxOnMercator(
							Square			&sq_in,
							Square			*psq_array_ret
)
{
	NUM_TYPE			x0;
	NUM_TYPE			y0; 
	NUM_TYPE			x1;
	NUM_TYPE			y1;
	NUM_TYPE			x_min;
	NUM_TYPE			x_max;
	NUM_TYPE			y_min;
	NUM_TYPE			y_max;
	static double
	const				PI2dev3 = 2.0*PI/3.0;
	Square				sq_sub1;
	Square				sq_sub2;
	Square				outersq;
	outersq = getOuterSquare( sq_in);// Max size of the box
	x0		= ( outersq.getVec( 0)).x();
	y0		= ( outersq.getVec( 0)).y(); 
	x1		= ( outersq.getVec( 2)).x();
	y1		= ( outersq.getVec( 2)).y();

	x_max	= getLarger( x0, x1);
	x_min	= getSmaller( x0, x1);
	y_max	= getLarger( y0, y1);
	y_min	= getSmaller( y0, y1);	

	if( fabs( getLarger( x0, x1) - getSmaller( x0, x1)) >= PI2dev3)
	{
		sq_sub1.iassign( 0, x_max,	y_max, 0);	// left up
		sq_sub1.iassign( 1, PI,		y_max, 0);	// right up
		sq_sub1.iassign( 2, PI,		y_min, 0);	// right down
		sq_sub1.iassign( 3, x_max,	y_min, 0);	// left down

		psq_array_ret[0] = sq_sub1;
		//-------------------------------------------------------------------
		//	If x_min == x_max, this won't produce 2 squares. only 1.
		//-------------------------------------------------------------------
		if( fabs( x_max) - fabs( x_min) < 0.000001)
			return 1;
		sq_sub2.iassign( 0, -PI,	y_max, 0);	// left up
		sq_sub2.iassign( 1, x_min,	y_max, 0);	// right up
		sq_sub2.iassign( 2, x_min,	y_min, 0);	// right down
		sq_sub2.iassign( 3, -PI,	y_min, 0);	// left down
		psq_array_ret[1] = sq_sub2;
		return 2;
	}
	else
	{
		psq_array_ret[0] = outersq;
		return 1;
	}
} 

//---------------------------------------------------------------------------
//	extractSurface()
//---------------------------------------------------------------------------
bool Sankaku::blitSurface(


						SDL_Surface*	&surface_in,
						NUM_TYPE		x0,
						NUM_TYPE		y0,
						NUM_TYPE		ext_width,
						NUM_TYPE		ext_height,
						SDL_Surface*	&surface_out,
						int				destx,
						SDL_Rect		*rect_copied	// destination area
)
{
	SDL_Rect area;
	SDL_Rect destrect;
	area.x = fixPos( round( x0), surface_in->w);
	area.y = fixPos( round( y0), surface_in->h);
	area.w = round( ext_width);
	area.h = round( ext_height);
	//-----------------------------------------------------------------------
	//	destrect.w and .h doesn't need to be initialize since
	//	SDL_BlitSurface() doesn't care those values.
	//-----------------------------------------------------------------------	
    destrect.x = destx;// doubtable!
	destrect.y = 0;
	//-----------------------------------------------------------------------
	//	area.x and .y must be less than surface_in->w
	//-----------------------------------------------------------------------
	if( area.x + area.w > surface_in->w)	area.w = surface_in->w - area.x;
	if( area.y + area.h > surface_in->h)	area.h = surface_in->h - area.h;
	
	if( SDL_BlitSurface( surface_in, &area, surface_out, &destrect) != 0)
		return false;
	pcallback->reportBox( area, current_trig);
	//-----------------------------------------------------------------------
	//	Set return value which is the destination rectangle
	//-----------------------------------------------------------------------
	rect_copied->w = area.w;		//	SDL_Blit..() doesn't change destrect
	rect_copied->h = area.h;		//	So we have to do it ourselves
	rect_copied->x = destrect.x;	//	
	rect_copied->y = 0;
	return true;
}


//---------------------------------------------------------------------------
//			extractSurfaceFromMercator()
//
//	transfer sq on mercator to psurface_out
//---------------------------------------------------------------------------
bool Sankaku::extractSurfaceFromMercator(
							Square			sq,
							SDL_Surface*	&psurface_out,
							int				destx_max,
							NUM_TYPE		destx,// theoretical coord val
							SDL_Rect		*rect_copied
)
{
	int			fixed_dstx = destx_max;
	int			ideal_dstx = round( destx * FACTOR_W);
	NUM_TYPE	final_x, final_y, final_w, final_h;// area where copied from
	//-----------------------------------------------------------------------
	// Obtaining lower left point and upper right point. note that sq holds
	// coordinates counter-clockwise start from lower left.
	//-----------------------------------------------------------------------
	NUM_TYPE x0 = getSmaller(	( sq.getVec( 0)).x(), ( sq.getVec( 2)).x());
	NUM_TYPE y0 = getSmaller(	( sq.getVec( 0)).y(), ( sq.getVec( 2)).y());
	NUM_TYPE x1 = getLarger(	( sq.getVec( 0)).x(), ( sq.getVec( 2)).x());
	NUM_TYPE y1 = getLarger(	( sq.getVec( 0)).y(), ( sq.getVec( 2)).y());
	//-----------------------------------------------------------------------
	//	Fix the theoretical coordinate to actual coordinate.
	//	But still double for now.
	//-----------------------------------------------------------------------
	final_x = sfc_mercator->w / 2.0 + FACTOR_W * x0;
	final_y = FACTOR_H * ( 1.0 - y0);
	final_w = FACTOR_W * ( x1-x0);
	final_h = FACTOR_H * ( y1-y0);
	final_y -= final_h;
	if( final_w < 1.0)	final_w = 1.0;
	if( final_h < 1.0)	final_h = 1.0;
	//-----------------------------------------------------------------------
	//	This is important. 
	//-----------------------------------------------------------------------
	if( ideal_dstx < destx_max)		fixed_dstx = ideal_dstx;
	//-----------------------------------------------------------------------
	//	Call blitSurface() to blit it.
	//-----------------------------------------------------------------------
	if( !blitSurface(
							sfc_mercator,
							final_x,

							final_y,
							final_w,
							final_h,
							psurface_out,
							fixed_dstx,
							rect_copied
						)
	)
		return false;
	rect_copied->w = rect_copied->w - ( destx_max - fixed_dstx);

	return true;
}

//---------------------------------------------------------------------------
//				drawProcessingBox()
//
//	Copy rectangle area from sfc_mercator to sfc_progress.
//	The area on sfc_mercator is given by theoretical value sq_in.
//	If the rectangle area specified by sq_in over edge to another side of
//	edge, devide the area into 2 piece for each edge and copy them to
//	sfc_progress.
//	Then draw coloed box on the copied area on sfc_mercator.
//
//	sq_in		: Theoretical coordinated area on sfc_mercator
//	red			: Color component of Box R of RGB
//	green		: Color component of Box G of RGB
//	blue		: Color component of Box B of RGB
//---------------------------------------------------------------------------
SDL_Rect Sankaku::drawProcessingBox(
	Square &sq_in
)
{
	Square			sq_array[2];
	unsigned int	num_sqs;
	SDL_Rect		dstrect;
	SDL_Rect		retrect;
	SDL_Rect		rect_copied1;
	SDL_Rect		rect_copied2;
	NUM_TYPE		theo_dstx = 0.0;

	rect_copied1.x = rect_copied1.y = rect_copied1.w = rect_copied1.h = 0;
	rect_copied2.x = rect_copied2.y = rect_copied2.w = rect_copied2.h = 0;
	//-----------------------------------------------------------------------
	//	Obtain Number of Squares and Those Coordinates
	//-----------------------------------------------------------------------
	num_sqs		= getBoxOnMercator( sq_in, sq_array);
	dstrect.w	= sfc_progress->w;
	dstrect.h	= sfc_progress->h;
	dstrect.x	= 0;
	dstrect.y	= 0;
	//-----------------------------------------------------------------------
	//	If there is 2nd box, draw it first.
	//-----------------------------------------------------------------------
	if( num_sqs >= 2)
	{	//-------------------------------------------------------------------
		//	Copy the rectangle area to the sfc_progress.
		//-------------------------------------------------------------------
		if( !extractSurfaceFromMercator(
											sq_array[ 1],
											sfc_progress,
											0,		// dextx max real coord.
											0.0,	// destx theoretical
											&rect_copied2
										)
		)
		{
			cerr << "can't extract in drawProcessingBox()" << endl;
			return dstrect;
		}
		theo_dstx = sq_array[1].getVec( 2).x() - sq_array[1].getVec( 0).x(); 
	}
	//-----------------------------------------------------------------------
	//	Copy the rectangle area to the sfc_progress.
	//-----------------------------------------------------------------------
	if( !extractSurfaceFromMercator(
										sq_array[ 0],
										sfc_progress,
										rect_copied2.w,
										theo_dstx,
										&rect_copied1
									)
	)
	{
		cerr << "can't extract in drawProcessingBox()" << endl;
		return dstrect;
	}
	//-----------------------------------------------------------------------
	//	Calculate the copied area
	//-----------------------------------------------------------------------
	retrect.x = 0;
	retrect.y = 0;
	retrect.w = rect_copied1.w + rect_copied2.w;
	//-----------------------------------------------------------------------
	//	This version is not very serious about y-axis. So just take lower
	//	height.
	//-----------------------------------------------------------------------
	if( num_sqs == 1)
		retrect.h = rect_copied1.h;
	else
		retrect.h =	(
						rect_copied1.h > rect_copied2.h
						?
						rect_copied2.h : rect_copied1.h
					);
	return retrect;
}

//---------------------------------------------------------------------------
//				extractSurfaceFromProcessing()
//
//	Copy and Resize rectanble area on sfc_progress to sfc_trig[ index_trig]
//
//	rect_in		: Rectangle area on sfc_progress which is copied from
//	sq			: Theoretical coordinate rectangle on triangle.
//				  Image will be copied to here.
//	index_trig	: Index of sfc_trig
//---------------------------------------------------------------------------
bool Sankaku::extractSurfaceFromProcessing(
	SDL_Rect &rect_in, Square &sq, unsigned int index_trig
)
{
	SDL_Rect		dstrect, rct_actual;
	SDL_Surface		*sfc_scaled		= NULL;
	SDL_Surface		*sfc_non_scaled	= NULL;
	string			err_msg;
	bool			retval			= true;
	// sq is Clockwise
	sq			= getOuterSquare( sq);
	Vec vec_ru	= sq.getVec( 1);
	Vec vec_ld	= sq.getVec( 3);
	//-----------------------------------------------------------------------
	//	Convert theoretical coordinates to actual coordinates
	//	Note that y axis is upside down
	//-----------------------------------------------------------------------
	dstrect.x = round( WIDTH_TRIANGLE * vec_ld.x());
	dstrect.y = round( sfc_trig_array[0]->h - WIDTH_TRIANGLE * vec_ld.y());
	dstrect.w = round( WIDTH_TRIANGLE * (vec_ru.x() - vec_ld.x()));
	dstrect.h = round( WIDTH_TRIANGLE * (vec_ru.y() - vec_ld.y()));
	dstrect.y -= dstrect.h;
	//-----------------------------------------------------------------------
	//	Fix the values
	//-----------------------------------------------------------------------
	if( dstrect.y < 0)	dstrect.y = 0;

	try{
		//-------------------------------------------------------------------
		//	Create New Surfaces
		//-------------------------------------------------------------------
		if( !createSurface( sfc_non_scaled, rect_in.w, rect_in.h, true))
		{
			err_msg = "create new surface";
			throw &err_msg;
		}
		//-------------------------------------------------------------------
		//	Blit it from sfc_progree to sfc_non_scaled
		//-------------------------------------------------------------------
        SDL_BlitSurface( sfc_progress, &rect_in, sfc_non_scaled, NULL);
		//-------------------------------------------------------------------
		//	Scale sfc_non_scaled to fit sfc_scaled
		//-------------------------------------------------------------------
		if( dstrect.w < min_pix)		dstrect.w = min_pix;
		if( dstrect.h < min_pix)		dstrect.h = min_pix;
		//-------------------------------------------------------------------
		//	Resize the obtained surface.
		//-------------------------------------------------------------------
		if( !transformSurfaceEx(
									sfc_non_scaled,
									sfc_scaled,
									dstrect.w,		// Width to be scaled
									dstrect.h,		// Height to be scaled
									&rct_actual,
									true
								)
		)
		{
			err_msg	= "transform surface";
			throw &err_msg;
		}
    	rect_in.x = 0; rect_in.y = 0;
		//-------------------------------------------------------------------
		//	Blit it from sfc_scaled to triangle's surface.
		//	sfc_progress must be cleared with black for the transparent
		//	reason.
		//-------------------------------------------------------------------
		SDL_BlitSurface(
						sfc_scaled,
						&rct_actual,
						sfc_trig_array[ index_trig],
						&dstrect
					);
		if( SDL_FillRect( sfc_progress, &rct_actual, 0x000000) == -1)
		{
			cout << "SDL_FillRect() Failed" << endl;
		}
		retval = true;
	}catch( string *msg)
	{
		if( !retval)
			cerr << "failed to " << *msg
				 << " in extractSurfaceFromProcessing()\n";
		retval = false;
	}
	freeSurface( sfc_non_scaled);
	freeSurface( sfc_scaled);
	return retval;
}



//---------------------------------------------------------------------------
//	drawLine4Tab()
//---------------------------------------------------------------------------
void Sankaku::drawLine4Tab( Square &sq)
{
	unsigned int x0, y0, x1, y1, counter;

	for( counter = 0; counter < 3; counter ++)
	{
		x0 = round( ( sq.getVec( counter	)).x());
		y0 = round( ( sq.getVec( counter 	)).y());
		x1 = round( ( sq.getVec( counter + 1)).x());
		y1 = round( ( sq.getVec( counter + 1)).y());
		sge_Line( sfc_result, x0, y0, x1, y1, 0x000000);
	}
}

//---------------------------------------------------------------------------
//	drawTab()
//---------------------------------------------------------------------------
void Sankaku::drawTab()
{
	Square sq1, sq2, sq3;
	int space = WIDTH_TRIANGLE / 10;
	sq1.iassign( 0, 2 * space,				2 * space,		0);
	sq1.iassign( 1, 3 * space,				space,			0);
	sq1.iassign( 2, space + WIDTH_TRIANGLE, space,			0);
	sq1.iassign( 3, 2 * space + WIDTH_TRIANGLE, 2 * space,	0);
	drawLine4Tab( sq1);

	sq2.iassign( 0, sq1.getVec(1).x(),	sq1.getVec(1).y(),	0);
	sq2.iassign( 1, space,				3 * space,			0);
	sq2.iassign( 2,
				 2 * space + WIDTH_TRIANGLE / 2 - space * sqrt( 2.0),
				 2 * space + HEIGHT_TRIANGLE,
				 0
				);
	sq2.iassign( 3,
				 2 * space + WIDTH_TRIANGLE / 2,
				 2 * space + HEIGHT_TRIANGLE,
				 0
				);
	drawLine4Tab( sq2);

	sq3.iassign( 0, sq2.getVec(3).x(),			sq2.getVec(3).y(),		0);
	sq3.iassign( 1, sq3.getVec(0).x()+space,	sq3.getVec(0).y()+space,0);
	sq3.iassign( 2, space+WIDTH_TRIANGLE*3/2,	3*space+HEIGHT_TRIANGLE,0);
	sq3.iassign( 3, 2*space+WIDTH_TRIANGLE*3/2,	2*space+HEIGHT_TRIANGLE,0);
	drawLine4Tab( sq3);	
}

//---------------------------------------------------------------------------
//	unfold()
//---------------------------------------------------------------------------
bool Sankaku::unfold()
{
	int	space	= WIDTH_TRIANGLE / 5;
	//-------------------------------------------------------------
	// Transfer trig0 to sfc_result
	//-------------------------------------------------------------
	sge_transform(  // First, rotate trig0 to sfc_temp_unfold
				sfc_trig_array[ 0], 
				sfc_temp_unfold, 
				60.0,							// angle 
				1.0, 1.0,						// scale 
				(sfc_trig_array[0])->w / 2,     // pivot point
				(sfc_trig_array[0])->h * 2 / 3, 
				(sfc_trig_array[0])->w / 2,		// destination point
				(sfc_trig_array[0])->h * 1 / 3,
				0
				);
	sge_BlitTransparent(
				sfc_temp_unfold,
				sfc_result,
				0,	0,
				space,	space,
				WIDTH_TRIANGLE, HEIGHT_TRIANGLE,
				0xffffff,
				255
				);
	//-------------------------------------------------------------
	// Transfer trig1 to sfc_result
	//-------------------------------------------------------------
	sge_BlitTransparent(
				sfc_trig_array[ 1],
				sfc_result,
				0,	0,
				space + WIDTH_TRIANGLE / 2,	space,
				WIDTH_TRIANGLE,				HEIGHT_TRIANGLE,
				0xffffff,
				255
				);
	//-------------------------------------------------------------
	// Transfer trig2 to sfc_result
	//-------------------------------------------------------------
	// First, clear sfc_temp_unfold.
	// Then rotate trig2 to sfc_temp_unfold
	sge_ClearSurface( sfc_temp_unfold, 0xff, 0xff, 0xff);
	// rotate
	sge_transform(
				sfc_trig_array[ 2], 
				sfc_temp_unfold, 
				-60.0,							// angle 
				1.0, 1.0,						// scale 
				(sfc_trig_array[0])->w / 2,
				(sfc_trig_array[0])->h * 2 / 3, 
				(sfc_trig_array[0])->w / 2,
				(sfc_trig_array[0])->h * 1 / 3,	// dest point 
				0
				);
	// Then transfer sfc_temp_unfold to sfc_result
	sge_BlitTransparent(
				sfc_temp_unfold,
				sfc_result,
				0,	0,
				space + WIDTH_TRIANGLE,	space,
				WIDTH_TRIANGLE,			HEIGHT_TRIANGLE,
				0xffffff,
				255
				);
	//-------------------------------------------------------------
	// Transfer trig3 to sfc_result
	//-------------------------------------------------------------
	sge_BlitTransparent(
				sfc_trig_array[ 3],
				sfc_result,
				0,	0,
				space + WIDTH_TRIANGLE * 3 / 2 - 1,	space,
				WIDTH_TRIANGLE,						HEIGHT_TRIANGLE,
				0xffffff,
				255
				);
	drawTab();
	return true;
}

//---------------------------------------------------------------------------
//	checkSfcProgSize()
//	Check if the sfc_progress is right size or not. Because the size
//	can't be predicted untill sfc_mercator is normalized.
//	If check failed, regenerate sfc_progress.
//---------------------------------------------------------------------------
bool Sankaku::checkProgSfcSize()
{
	int		rightw, righth;
	bool	change = false;
	//-----------------------------------------------------------------------
	//	Check if we loaded everything
	//-----------------------------------------------------------------------
	if( sfc_mercator == NULL)
	{
		cout << "sfc_mercator is not initialized" << endl;
		return false;
	}
	if( sfc_progress == NULL) createSurface(	sfc_progress, 100, 100, true );
	//-----------------------------------------------------------------------
	//	Obtain w and h temporarity
	//-----------------------------------------------------------------------
	rightw = sfc_progress->w;
	righth = sfc_progress->h;
	//-----------------------------------------------------------------------
	//	If wrong size, set flag to scale it
	//-----------------------------------------------------------------------
	if( sfc_progress->w < sfc_mercator->w)
	{
		rightw = sfc_mercator->w;
		change = true;
	}
	if( sfc_progress->h < sfc_mercator->h / 5)
	{
		righth = sfc_mercator->w / 5;
		change = true;
	}
	//-----------------------------------------------------------------------
	//	If flag is set. Free original one and create new right sized one.
	//-----------------------------------------------------------------------
	if( change)
	{
		SDL_FreeSurface( sfc_progress);
		//-----------------------------------------------------------------------
		//	Create Surface sfc_progress
		//-----------------------------------------------------------------------
		if( !createSurface(	sfc_progress, rightw, righth, true))
		{
			cerr << "Unable to create sfc_progress" << endl;
			return false;
		}
	}
	return true;
}
//---------------------------------------------------------------------------
//	initProgram()
//---------------------------------------------------------------------------
bool Sankaku::initSankaku( unsigned int reso, unsigned int sz)
{
	unsigned int counter;
	//-----------------------------------------------------------------------
	//	If it's not ready yet, free everything to be ready.
	//-----------------------------------------------------------------------
	if( ready2play)
	{
		freeEmAllButMercator();// also ready2play set to false in this func.
		nullEmAllButMercator();
	}
	//-----------------------------------------------------------------------
	//	If call back function is not set, return false
	//-----------------------------------------------------------------------
	if( pcallback == NULL)
	{
		cerr << "call back is not initialized" << endl;
		return false;
	}
	//-----------------------------------------------------------------------
	//	If the sfc_mercator is not loaded, return error.
	//-----------------------------------------------------------------------
	if( sfc_mercator == NULL)
	{
		cerr << "File is not loaded" << endl;
		return false;
	}
	//-----------------------------------------------------------------------
	//	Initializing and Obtaining HEIGHT_TRIANGLE
	//-----------------------------------------------------------------------
	RESOLUTION		= 1.0 / double( reso);
	WIDTH_TRIANGLE	= sz;
	initPlanet( WIDTH_TRIANGLE, HEIGHT_TRIANGLE, RESOLUTION);
	min_pix			= roundUp( WIDTH_TRIANGLE * RESOLUTION);
	//-----------------------------------------------------------------------
	//	Create Surface sfc_trig_array
	//-----------------------------------------------------------------------
	for( counter = 0; counter < 4; counter++)
	{
		if( !createSurface(
							sfc_trig_array[ counter],
							WIDTH_TRIANGLE,
							HEIGHT_TRIANGLE,
							false
						)
		)
		{
			cerr << "Unable to create sfc_trig_array[]" << endl;
			return false;
		}
	}
	//-----------------------------------------------------------------------
	//	Create Surface sfc_progress
	//-----------------------------------------------------------------------
	if( !createSurface(	sfc_progress, 100, 100, false))
	{
		cerr << "Unable to create sfc_progress" << endl;
		return false;
	}
	//-----------------------------------------------------------------------
	//	Create Surface sfc_result
	//-----------------------------------------------------------------------
	if( !createSurface(
						sfc_result,
						round( 2.9 * WIDTH_TRIANGLE),//1/5+2.5+1/5=2.9
						round( 1.4 * HEIGHT_TRIANGLE), false
					)
	)
	{
		cerr << "Unable to create sfc_result" << endl;
		return false;
	}
	//-----------------------------------------------------------------------
	//	Create Surface sfc_temp_unfold
	//-----------------------------------------------------------------------
	if( !createSurface(sfc_temp_unfold,WIDTH_TRIANGLE,HEIGHT_TRIANGLE,false))
	{
		cerr << "Unable to create sfc_temp_unfold" << endl;
		return false;
	}
	//-----------------------------------------------------------------------
	// clear surfaces with white
	//-----------------------------------------------------------------------
	for( counter = 0; counter < 4; counter++)
		sge_ClearSurface( sfc_trig_array[ counter], 255, 255, 255);
	sge_ClearSurface( sfc_result, 255, 255, 255);
	sge_ClearSurface( sfc_temp_unfold, 255, 255, 255);
	//-----------------------------------------------------------------------
	// Check if sfc_progress is right size. If not, renew it.
	//-----------------------------------------------------------------------
	if( !checkProgSfcSize())
	{
		cerr << "sfc_progress regeneration failed" << endl;
		return false;
	}
	//-----------------------------------------------------------------------
	//	Set flag to show redy
	//-----------------------------------------------------------------------
	ready2play =	true;
	return			true;
}

//---------------------------------------------------------------------------
//	Sankaku::normalizeMercator()
//	Reshape the loaded sfc_mercator to ( h : w = 1 : PI)
//---------------------------------------------------------------------------
bool Sankaku::normalizeMercator()
{
	SDL_Surface		*sfc_normalized = NULL;
	SDL_Rect		rect_normalized;
	double			error_x,	error_y;
	int				fixed_x,	fixed_y;
	int				actual_x,	actual_y;
	int				correct_x,	correct_y;
	
	actual_x = sfc_mercator->w;
	actual_y = sfc_mercator->h;
	//-----------------------------------------------------------------------
	//	 If rate of side ( w : h) = ( PI : 1), just return
	//-----------------------------------------------------------------------
	if( actual_x == round( PI * actual_y)|| actual_y == round( actual_x/PI))
		return true;
	//-----------------------------------------------------------------------
	//	If not, adjust the size
	//	x:y=PI:1 => x=PI*y or y = x/PI	
	//	If x is fixed, y = x/PI
	//	If y is fixed, x = y*PI
	//-----------------------------------------------------------------------
	correct_y	= round( float( actual_x) / PI);  
	error_y		= double( abs( int(actual_y - correct_y)))
				/ double( correct_y);
	correct_x	= round( float( actual_y) * PI);
	error_x		= double( abs( int(actual_x - correct_x)))
				/ double( correct_x);
	if( error_x < error_y)// take smaller error one, = say x is fixed
	{
		fixed_x = actual_x;
		fixed_y = correct_y;
	}
	else
	{
		fixed_x = correct_x;
		fixed_y = actual_y;
	}	
	//-----------------------------------------------------------------------
	//	  Scale into right ratio and size
	//-----------------------------------------------------------------------
	if( !transformSurfaceEx(
								sfc_mercator,
								sfc_normalized,
								fixed_x,
								fixed_y,
								&rect_normalized,true
							)
	)	
	{
		cerr << "transformSurface failed" << endl;
		SDL_FreeSurface( sfc_normalized);
		return false;
	}
	//-----------------------------------------------------------------------
	//	 Discard wrong sized surface and Assign new one
	//-----------------------------------------------------------------------
	SDL_FreeSurface( sfc_mercator);
	sfc_mercator = sfc_normalized;
	cout	<< "Image is resized from ( "
			<< actual_x				<< ", "
			<< actual_y				<< ") to ( "
			<< rect_normalized.w	<< ", " 
			<< rect_normalized.h	<< ")";

	return true;
}

//---------------------------------------------------------------------------
//	Sankaku::load()
//	Load bitmap to sfc_mercator and resize it by calling normalizeMercator()
//---------------------------------------------------------------------------
bool Sankaku::load( string *pstr_filename_in)
{
	freeSurface( sfc_mercator);
 	sfc_mercator = IMG_Load_RW( 
							SDL_RWFromFile(pstr_filename_in->c_str(), "rb"),
							 0
					); 
	if( sfc_mercator == NULL)
	{
		cerr	<< "Failed to Load Bitmap File \""	<< *pstr_filename_in 
				<< "\" : " << IMG_GetError()		<< endl;
	 	return false;
	}
	normalizeMercator();
	FACTOR_W	= sfc_mercator->w / ( 2 * PI);
	FACTOR_H	= sfc_mercator->h / 2.0;
	return true;
}


//---------------------------------------------------------------------------
//	Sankaku::save()
//	Save sfc_result
//---------------------------------------------------------------------------
bool Sankaku::save( string *pstr_filename_out)
{
	if( sfc_result == NULL)
	{
		cerr << "File is not treated yet" << endl;
		return false;
	}
	if( SDL_SaveBMP( sfc_result, pstr_filename_out->c_str()) == -1)
		return false;
	return true;
}


//---------------------------------------------------------------------------
//	Sankaku::play()
//	Main calculation loop
//---------------------------------------------------------------------------
bool Sankaku::play()
{
	//-----------------------------------------------------------------------
	//	sq[0] is square on p1 projected onto Mercator
	//	sq[1] is square on p2 ...
	//-----------------------------------------------------------------------
	Square			sq[4];
	Square			sq_p0;
	double			counter_y	= 0;
	double			counter_x	= 0;
	double			y_max		= sqrt( 3.0) / 2.0;
	double			x_max		= 1 - counter_y * sqrt( 3.0) / 3.0;
	unsigned int	milage		= 0;
	unsigned int	milage_show	= 0;
	unsigned int	milage_old	= 0;
	const unsigned int way2go	= int(
										sqrt( 3.0)	/
										4.0			/
										( RESOLUTION * RESOLUTION)
									);
    unsigned int	counter_i;
	SDL_Rect		rectonprogress;
	//-----------------------------------------------------------------------
	//	If it's not ready, return false
	//-----------------------------------------------------------------------
	if( !ready2play)	return false;
	//-----------------------------------------------------------------------
	//	Show some Info.
	//-----------------------------------------------------------------------
	cout << "RESOLUTION = " << RESOLUTION << endl;
	//-----------------------------------------------------------------------
	//	Matrix calculation part throw Error object if there is error
	//-----------------------------------------------------------------------
	try{
		while( counter_y < y_max)
		{
			counter_x = counter_y / sqrt( 3.0);
			while( counter_x < x_max)
			{
				if( stop)	break;
				//-----------------------------------------------------------
				//	Count how far we get in percent.
				//-----------------------------------------------------------
				milage_show = 100 * milage / way2go;
				if( milage_show > milage_old && milage_show < 100)
				{
					pcallback->reportPercent( milage_show);
					milage_old = milage_show;
				}
				//-----------------------------------------------------------
				//	Obtain square of p0
				//-----------------------------------------------------------
				getSquareOnTriangle( counter_x, counter_y, sq_p0);
				//-----------------------------------------------------------
				//	Obtain squares on p1, p2, p3 and p4 triangle
				//-----------------------------------------------------------
				sq[0] = sqcnv2p1( sq_p0);		// square of p1 onto Mercator
				sq[1] = sqcnv2p2fromp1( sq[0]);	// square of p2 onto Mercator
				sq[2] = sqcnv2p3fromp1( sq[0]);	// square of p3 onto Mercator
				sq[3] = sqcnv2p4( sq_p0);		// square of p4 onto Mercator
				//-----------------------------------------------------------
				//	Grab Image in the squares p1..p4 on Mercator and
				//	resize, and blit to triangle
				//-----------------------------------------------------------
				for( counter_i = 0; counter_i < 4; counter_i++)
				{
					current_trig = counter_i;
					// convert to spherical coordinate system
					sq[counter_i] = sqcartesian2Spherical(	sq[counter_i]);
					// convert to cylindrical projection
					sq[counter_i] = sqspherical2Mercator(	sq[counter_i]);
					// extruct rectangle from cylindrical projection image
					// onto progress surface, then obtain size of rectangle
					rectonprogress = drawProcessingBox(	sq[ counter_i]);
					// resize the rectangle and draw it on triangle
					if( !extractSurfaceFromProcessing(
						rectonprogress, sq_p0, counter_i
					))
					{
						cerr	<< "Failed to extract image "
								<< "from progress surface" << endl;
						return false;
					}
				}
				counter_x += RESOLUTION;
				milage++;
			}
			counter_y += RESOLUTION;
		}
	}
	catch( Error er){
		cerr << "Error in Engine";
	}

	//-----------------------------------------------------------------------

	//	transfer 4 triangles into unfolded diagram
	//-----------------------------------------------------------------------
	cout << "unfolding..." << endl;
	unfold();
	cout << "done" << endl;
	//-----------------------------------------------------------------------
	//	Report that All Done
	//-----------------------------------------------------------------------
	pcallback->reportPercent( 100);
	ready2play = false;

	return true;
}

//---------------------------------------------------------------------------
//	Sankaku::getMercatorSurface()
//---------------------------------------------------------------------------

SDL_Surface *Sankaku::getMercatorSurface() const
{
	return sfc_mercator;
}

//---------------------------------------------------------------------------
//	Sankaku::getResultSurface()
//---------------------------------------------------------------------------
SDL_Surface *Sankaku::getResultSurface() const
{
	return sfc_result;
}

//---------------------------------------------------------------------------
//	freeSurface()
//	Free SDL_Surface and set NULL
//---------------------------------------------------------------------------
void freeSurface( SDL_Surface* &sfc)
{
	if( sfc != NULL)
	{
		SDL_FreeSurface( sfc);
		sfc = NULL;
	}
}

//---------------------------------------------------------------------------
//	transformSurface()
//	Transform surface one to another and also takes care of scalling.
//	original	: source surface
//	newsfc		: newly created resized surface
//	prct_result	: size of dst. x and y are ignored
//---------------------------------------------------------------------------
bool transformSurfaceEx(
						SDL_Surface *original,
						SDL_Surface *&newsfc,
						int			sz_w,
						int			sz_h,
						SDL_Rect	*prct_result,
						bool		exact
){
	double	factorx = double( sz_w) / original->w;;
	double	factory = double( sz_h) / original->h;;
	double	adjustx = 0.05;
	double	adjusty = 0.05;
	int		counter = 0;

	//-----------------------------------------------------------------------
	//	newsfc must be empty. If newsfc already assigned, free it.
	//-----------------------------------------------------------------------
	if( newsfc != NULL)	freeSurface( newsfc);
	//-----------------------------------------------------------------------
	//	Create scaled surface. Loop continues untill newly created surface
	//	gets right size.
	//-----------------------------------------------------------------------
	while( true)
	{
		newsfc = PG_Draw::ScaleSurface (
										original,
										factorx,
										factory,
										false	// NO Smoothing
									);
		//-------------------------------------------------------------------
		//	Ending Condition
		//-------------------------------------------------------------------
		if( !exact)										break;

		if( sz_w <= newsfc->w && sz_h <= newsfc->h)		break;
		//-------------------------------------------------------------------
		//	Increase Scalling Factor
		//-------------------------------------------------------------------	
		if( sz_w > newsfc->w)
		{
			factorx = factorx * ( 1.0 + adjustx);
			adjustx++;
		}
		if( sz_h > newsfc->h)
		{
			factory = factory * ( 1.0 + adjusty);
			adjusty++;
		}
		//-------------------------------------------------------------------
		//	Emergency Stop. Will be erased in the Future
		//-------------------------------------------------------------------
		counter++;
		if( counter > 50)
		{
			cout << "Emergency Stop : Could not create a surface : ";
			cout << "sz_w = " << sz_w << ", sz_h = " << sz_h << endl;
			SDL_FreeSurface( newsfc);
			return false;
		}
		SDL_FreeSurface( newsfc);
	}
	prct_result->x = 0;
	prct_result->y = 0;
	prct_result->w = newsfc->w;
	prct_result->h = newsfc->h;

	return true;
}

/*// For Debug
#ifndef	_DEBUG
#include <cstdlib>

// For Debug
void printRect( char *name, SDL_Rect *rect, bool nl)
{
	cout << name << "= ( " << rect->x << ", " << rect->y << ", " << rect->w << ", " << rect->h << ")";
	if( nl)	cout << endl;
}
void printRect( char *name, SDL_Surface *sfc, bool nl)
{
	cout << name << "= ( " << sfc->w << ", " << sfc->h << ")";
	if( nl)	cout << endl;
}


void saveSurfacePieces( char *name, SDL_Surface *sfc)
{
	static int i = 0;
	char index[10];
	myitoa( i++, index, 10);
	string fn( name);
	fn.append( index);
	fn.append( ".bmp");
	SDL_SaveBMP( sfc, fn.c_str());
}

int myitoa( int d, char *string, int base)
{
	if(string) {
		return sprintf(string, "%d", d);
	}
	return -1;
}
#endif*/




