/***************************************************************************
                          sankakuw.cpp  -  description
                             -------------------
    begin                : Thu Nov 21 2002
    copyright            : (C) 2002 by Nobuhiro Kuroiwa
    email                : nobu@iastate.edu
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "SDL/SDL_thread.h"
#include "sankakuw.h"
#include "sankaku.h"
#include "paragui/pgdraw.h"

SankakuWindow::SankakuWindow( )
{
	pbtn_go				= NULL;	pbtn_exit	= NULL;
	pbtn_load			= NULL;	pbtn_save	= NULL;
	pedt_load			= NULL;	pedt_save	= NULL;
	pedt_reso			= NULL;	pedt_size	= NULL;
	plbl_reso			= NULL;	plbl_size	= NULL;
	pimg_mercator		= NULL; pimg_result	= NULL;

	sfc_mercator2show	= NULL;
	sfc_result2show		= NULL;

	colour[0]	= 0xFF0000;		colour[1]	= 0x00FF00;
	colour[2]	= 0x0000FF;		colour[3]	= 0xFFFF00;
	psankaku	= NULL;
	thread		= NULL;

	ready2go = initSankakuWindow();
}

//---------------------------------------------------------------------------
//	Sankaku::~SankakuWindow()
//	Destructor of SankakuWindow Class
//	Free Everything. Note that sfc_mercator2show and sfc_result2show are
//	freed by PG_Image class since the constructor's parameter set to true
//	which is default value.
//---------------------------------------------------------------------------
SankakuWindow::~SankakuWindow()
{
	if( psankaku		!= NULL)	psankaku->stopIt();

	SDL_Delay( 100);
	if( thread			!= NULL)	SDL_WaitThread( thread, NULL);
	/*if( pimg_mercator	!= NULL)	delete pimg_mercator;
	if( pimg_result		!= NULL)	delete pimg_result;
	if( pbtn_go			!= NULL)	delete pbtn_go;
	if( pbtn_exit		!= NULL)	delete pbtn_exit;
	if( pbtn_load		!= NULL)	delete pbtn_load;
	if( pbtn_save		!= NULL)	delete pbtn_save;
	if( pedt_load		!= NULL)	delete pedt_load;
	if( pedt_save		!= NULL)	delete pedt_save;
	if( pedt_reso		!= NULL)	delete pedt_reso;
	if( pedt_size		!= NULL)	delete pedt_size;
	if( plbl_reso		!= NULL)	delete plbl_reso;
	if( plbl_size		!= NULL)	delete plbl_size;*/

	if( psankaku		!= NULL)	delete psankaku;

}

void SankakuWindow::reportBox( SDL_Rect rect, unsigned short trig_index)
{
	static unsigned int counter = 0;
	int x	= int( magnification * rect.x);
	int y	= int( magnification * rect.y);
	int w	= int( magnification * ( rect.x + rect.w));
	int h	= int( magnification * ( rect.y + rect.h));

	x = fixPos( x, sfc_mercator2show->w);
	y = fixPos( y, sfc_mercator2show->h);
 
	sge_Rect( sfc_mercator2show, x, y, w, h, colour[ trig_index]);
  
	if( counter % 20 == 0)	pimg_mercator->Update();
	counter++;
}

void SankakuWindow::reportPercent( int per)
{
	cout << per << "% done" << endl;
	pimg_mercator->Update();
	if( per == 100)	showAllButtons();
}

bool SankakuWindow::isReady()
{
	return ready2go;
}

void SankakuWindow::show()
{
	showAllButtons();
	pedt_load->Show();
	pedt_save->Show();
	pedt_reso->Show();
	pedt_size->Show();
	plbl_reso->Show();
	plbl_size->Show();
	pimg_mercator->Show();
	pimg_result->Show();
}

void SankakuWindow::showAllButtons()
{
	pbtn_go->Show();
	pbtn_exit->Show();
	pbtn_load->Show();
	pbtn_save->Show();
}

void SankakuWindow::hideAllButtons()
{
	pbtn_load->Hide( true);
	pbtn_save->Hide( true);
	pbtn_go->Hide( true);
	pbtn_exit->Hide( true);
}
//---------------------------------------------------------------------------
//	initProgram()
//---------------------------------------------------------------------------
bool SankakuWindow::initSankakuWindow()
{
	pos_mercator.x		= 0; pos_mercator.y	=	0;
	pos_result.x		= 0; pos_result.y	=	200;
	//-----------------------------------------------------------------------	
	// Create Surfaces
	//-----------------------------------------------------------------------
	sfc_mercator2show	= SDL_CreateRGBSurface(
										SDL_SWSURFACE,
										628, 200,
										32,
										0, 0, 0, 0
									);
	sfc_result2show		= SDL_CreateRGBSurface(
										SDL_SWSURFACE,
										540, 240,
										32,
										0, 0, 0, 0
									);

	if( sfc_mercator2show == NULL || sfc_result2show == NULL)
	{
		freeSurface( sfc_mercator2show);
		freeSurface( sfc_result2show);
		return false;
	}
	//-----------------------------------------------------------------------	
	// Create Widgets
	//-----------------------------------------------------------------------	
	pimg_mercator = new PG_Image(
		NULL,	pos_mercator,	sfc_mercator2show,	true, "ThemeWidget"
	);

	pimg_result = new PG_Image(
		NULL,	pos_result,		sfc_result2show,	true,	"ThemeWidget"
	);
	PG_Rect rect( 540 + 10, 200 + 7, 80, 40);
	pbtn_load	= new PG_Button( NULL, 1, rect, "load");
	rect		= PG_Rect(	rect.x,rect.y + 7 + 40,	rect.w, rect.h);
	pbtn_save	= new PG_Button( NULL, 1, rect, "save");
	rect		= PG_Rect( rect.x, rect.y + 7 + 40,	rect.w, rect.h);
	pbtn_go		= new PG_Button( NULL, 1, rect, "go");

	rect		= PG_Rect( rect.x, rect.y + 7 + 40,	rect.w, rect.h);
	pbtn_exit	= new PG_Button( NULL, 1, rect, "exit");

	rect.w		= 40; rect.y = rect.y + 7 + 40;
	plbl_reso	= new PG_Label( NULL, rect, "RZ:");
	rect		= PG_Rect( rect.x + 40, rect.y,	rect.w, rect.h);
	pedt_reso	= new PG_LineEdit( NULL, rect);

	rect.x		-= 40;	rect.y	= rect.y + 7 + 40;
	plbl_size	= new PG_Label( NULL, rect, "SZ:");
	rect		= PG_Rect( rect.x + 40, rect.y ,	rect.w, rect.h);
	pedt_size	= new PG_LineEdit( NULL, rect);

	rect.x = 0; rect.y = 440; rect.w = 270; rect.h = 40;
	pedt_load	= new PG_LineEdit( NULL, rect);
	rect.x = 270; rect.y = 440;
	pedt_save	= new PG_LineEdit( NULL, rect);
	//-----------------------------------------------------------------------
	// this defines our callback handler for the message MSG_BUTTONCLICK,
	// we pass a pointer to the app object as userdata
	//-----------------------------------------------------------------------
	pbtn_go->SetEventObject(
		MSG_BUTTONCLICK,								this,
		(MSG_CALLBACK_OBJ)&SankakuWindow::go_handler,	NULL
	);
	pbtn_exit->SetEventObject(
		MSG_BUTTONCLICK,								this,
		(MSG_CALLBACK_OBJ)&SankakuWindow::exit_handler,	NULL
	);
	pbtn_load->SetEventObject(
		MSG_BUTTONCLICK,								this,
		(MSG_CALLBACK_OBJ)&SankakuWindow::load_handler,	NULL
	);
	pbtn_save->SetEventObject(	
		MSG_BUTTONCLICK,								this,
		(MSG_CALLBACK_OBJ)&SankakuWindow::save_handler,	NULL
	);
	//-----------------------------------------------------------------------	
	// Set Label to the Line Edit Widgets
	//-----------------------------------------------------------------------
	pedt_load->SetText( DEFAULT_FILE_IN);
	pedt_save->SetText( DEFAULT_FILE_OUT);
	pedt_reso->SetText( DEFAULT_RESOLUTION);
	pedt_size->SetText( DEFAULT_SIZE);
	return true;
}

//---------------------------------------------------------------------------
//	go_handler()
//---------------------------------------------------------------------------
PARAGUI_CALLBACK( SankakuWindow::go_handler)
{
	int 	reso,	size;
	string	errmsg;
	//-----------------------------------------------------------------------
	//	HIde all buttons except exit
	//-----------------------------------------------------------------------
	hideAllButtons();
	pbtn_exit->Show();

	cout << "go!" << endl;
	reso		= atoi( pedt_reso->GetText());
	size		= atoi( pedt_size->GetText());
	try
	{
		if( reso < 10 || size < reso)
		{
			errmsg.assign( "rezolution < 10 and size < resolution");
			throw &errmsg;
		}
		if( psankaku == NULL)
		{
			errmsg.assign( "File is not loaded");
			throw &errmsg;
		}
		if( !psankaku->initSankaku( reso, size))
		{
			errmsg.assign( "initialization failer");
			throw &errmsg;
		}
		if( !psankaku->isReady())
		{
			errmsg.assign( "Sankaku is not ready.");
			throw &errmsg;
		}
		//-----------------------------------------------------------------------
		//	Making ThumbNail
		//-----------------------------------------------------------------------
		if( !makeThumbNail())
		{
			errmsg.assign( "Thumbnail Loading Failed");
			throw &errmsg;
		}
		thread = SDL_CreateThread( SankakuWindow::letsDance,(void*)this);
		return true;
	}
	catch( string *msg)
	{
		cerr << *msg << endl;
		showAllButtons();
		return true;
	}
}

//---------------------------------------------------------------------------
//	exit_handler()
//---------------------------------------------------------------------------
PARAGUI_CALLBACK( SankakuWindow::exit_handler)
{
	//-----------------------------------------------------------------------
	// we can pass in some pointer to any userdata
	// (in this case we get a pointer to the application object)
	//-----------------------------------------------------------------------
	PG_Application* app = (PG_Application*) clientdata;
	//-----------------------------------------------------------------------
	// exit the application eventloop
	//-----------------------------------------------------------------------
	app->Quit();
	//-----------------------------------------------------------------------
	// return true to signal that we have processed this message
	//-----------------------------------------------------------------------
	return true;
}

//---------------------------------------------------------------------------
//	load_handler()
//---------------------------------------------------------------------------
PARAGUI_CALLBACK( SankakuWindow::load_handler) 
{
	string	str_file_in;
	bool	retvalue;
	//-----------------------------------------------------------------------
	//	Hide all buttons
	//-----------------------------------------------------------------------
	hideAllButtons();
	//-----------------------------------------------------------------------
	//	Get Filename and Load it
	//-----------------------------------------------------------------------
	str_file_in  = pedt_load->GetText();
	cout << "Loading ... ";
	//-----------------------------------------------------------------------
	//	Renew Sankaku object. Sankaku isn't designed to recycle
	//-----------------------------------------------------------------------
	if( psankaku != NULL)	psankaku->~Sankaku();
	psankaku = new Sankaku( &str_file_in);
	psankaku->setCallBack( this);
	//-----------------------------------------------------------------------
	//	Making ThumbNail
	//-----------------------------------------------------------------------
	if( !makeThumbNail())
	{
		retvalue = true;
		cerr << "Thumbnail Loading Failed" << endl;
	}
	else
	{
		retvalue = true;
		cout << " loaded" << endl;
	}
	//-----------------------------------------------------------------------
	//	Show all buttons
	//-----------------------------------------------------------------------
	showAllButtons();

	return retvalue;
}

//---------------------------------------------------------------------------
//	save_handler()
//---------------------------------------------------------------------------
PARAGUI_CALLBACK( SankakuWindow::save_handler)
{
	string	str_file_out = pedt_save->GetText();
	cout << "Saving " << str_file_out << " ... ";
	if( psankaku == NULL)
	{
		cerr << "Failed to save : Sankaku isn't ready." << endl;
		return true;
	}
	if( psankaku->save( &str_file_out))
		cerr << " saved" << endl;
	return true;
}

//---------------------------------------------------------------------------
//	makeThumbNail()
//
//	transfer and resize Sankaku::sfc_mercator to sfc_mercator2show
//	Then get magnification. 
//---------------------------------------------------------------------------
bool SankakuWindow::makeThumbNail()
{
	SDL_Rect	rect;
	SDL_Surface	*sfc_mer = NULL;
	SDL_Surface	*sfc_new = NULL;
	//-----------------------------------------------------------------------
	//	Check if sankaku object is created. Creation of Sankaku object and
	//	Loading file is occur at the same time.
	//-----------------------------------------------------------------------
	if( psankaku == NULL)	return false;
	//-----------------------------------------------------------------------
	//	Obtain the surface
	//-----------------------------------------------------------------------
	sfc_mer	= psankaku->getMercatorSurface();
	if( sfc_mer == NULL)
	{
		cout << "File loading Failed" << endl;
		return false;
	}
	//-----------------------------------------------------------------------
	//	Obtain Magnification Factor. This will be used later.
	//-----------------------------------------------------------------------
	magnification	= float( sfc_mercator2show->w) / sfc_mer->w;
	//-----------------------------------------------------------------------
	//	Clear the Surface
	//-----------------------------------------------------------------------
	SDL_FillRect( sfc_mercator2show, NULL, 0);
	//-----------------------------------------------------------------------
	//	Resize the surface.
	//-----------------------------------------------------------------------
	if( !transformSurfaceEx(
								sfc_mer,
								sfc_new,
								sfc_mercator2show->w,
								sfc_mercator2show->h,
								&rect
							)
	)
	{
		cout << "transformSurface Error" << endl;
		return false;
	}
	SDL_BlitSurface(
						sfc_new,			// source
						NULL,
						sfc_mercator2show,	// destination
						NULL
					);
	pimg_mercator->Update();
	SDL_FreeSurface( sfc_new);
	return true;
}

//---------------------------------------------------------------------------
//	SankakuWindow::letsDance()
//	Run as a thread
//---------------------------------------------------------------------------
int	SankakuWindow::letsDance( void *psanwin)
{
	SDL_Surface		*sfc_nr2show	= NULL;
	SDL_Rect		rect;
	SankakuWindow	*psw = ( SankakuWindow*)psanwin;

	cout << "Thread Launched" << endl;

	psw->psankaku->play();

	if( !transformSurfaceEx(
							psw->psankaku->getResultSurface(),
							sfc_nr2show,
							psw->sfc_result2show->w,
							psw->sfc_result2show->h,
							&rect
						)
	)
	{
		cerr << "loading result failure" << endl;
		return false;
	}

	SDL_BlitSurface( sfc_nr2show, NULL, psw->sfc_result2show, NULL);
	SDL_FreeSurface( sfc_nr2show);
	psw->pimg_result->Update();
	psw->thread = NULL;
	return 0;
}



